-- This software is copyright Kong Inc. and its licensors.
-- Use of the software is subject to the agreement between your organization
-- and Kong Inc. If there is no such agreement, use is governed by and
-- subject to the terms of the Kong Master Software License Agreement found
-- at https://konghq.com/enterprisesoftwarelicense/.
-- [ END OF LICENSE 0867164ffc95e54f04670b5169c09574bdbd9bba ]

local constants = require("kong.constants")
local meta = require("kong.enterprise_edition.meta")
local version = require("kong.clustering.compat.version")
local utils = require("kong.tools.utils")
local regex_router_migrate = require("kong.clustering.compat.regex_router_path_280_300")

local type = type
local ipairs = ipairs
local table_insert = table.insert
local table_sort = table.sort
local table_concat = table.concat
local gsub = string.gsub
local deep_copy = utils.deep_copy
local split = utils.split
local fmt = string.format

local ngx = ngx
local ngx_log = ngx.log
local INFO = ngx.INFO
local WARN = ngx.WARN
local null = ngx.null

local extract_major_minor = version.extract_major_minor
local parse_version = version.from_string


local _log_prefix = "[clustering] "

local function info(...)
  ngx_log(INFO, _log_prefix, ...)
end

local function warn(...)
  ngx_log(WARN, _log_prefix, ...)
end

local REMOVED_FIELDS = require("kong.clustering.compat.removed_fields")
local CLUSTERING_SYNC_STATUS = constants.CLUSTERING_SYNC_STATUS
local KONG_VERSION = assert(parse_version(meta.version))

local MIN_DP_VERSION
do
  local dp_version = constants.CLUSTERING_MIN_DATA_PLANE_VERSION
  if dp_version then
    MIN_DP_VERSION = assert(parse_version(dp_version))
  end
end


local EMPTY = {}


local _M = {}


local function err_plugin_set(...)
  local msg = fmt(...)
  return nil, msg, CLUSTERING_SYNC_STATUS.PLUGIN_SET_INCOMPATIBLE
end

local function err_plugin_version(...)
  local msg = fmt(...)
  return nil, msg, CLUSTERING_SYNC_STATUS.PLUGIN_VERSION_INCOMPATIBLE
end


local function check_plugin_compatibility(cp, dp)
  local cp_version = KONG_VERSION
  local dp_version = parse_version(dp.dp_version)

  for _, plugin in ipairs(cp.plugins_list) do
    if cp.plugins_configured[plugin.name] then
      local name = plugin.name
      local cp_plugin = cp.plugins_map[name]
      local dp_plugin = dp.dp_plugins_map[name]


      if not dp_plugin then
        if cp_plugin.version then
          return err_plugin_set("configured %s plugin %s is missing from " ..
                                "data plane", name, cp_plugin.version)
        end

        return err_plugin_set("configured %s plugin is missing from data plane",
                              name)
      end

      -- special case for 2.8 DP and 3.0/3.1 CP
      if cp_version.major == 3 and cp_version.minor <= 1 and
         dp_version.major == 2 and dp_version.minor == 8
      then
        goto continue
      end

      -- CP plugin needs to match DP plugins with major version
      -- CP must have plugin with equal or newer version than that on DP
      if cp_plugin.version and dp_plugin.version and
         (cp_plugin.major ~= dp_plugin.major or
          cp_plugin.minor < dp_plugin.minor)
      then
        return err_plugin_version("configured data plane %s plugin " ..
                                  "version %s is different from control " ..
                                  "plane version %s", plugin.name,
                                  dp_plugin.version, cp_plugin.version)
      end
    end

    ::continue::
  end

  return true, nil, CLUSTERING_SYNC_STATUS.NORMAL
end



local function kong_incompatible(dp, cp, extra)
  local msg = fmt("data plane version (%s) is incompatible with " ..
                  "control plane version (%s)", dp.string, cp.string)

  if extra then
    msg = msg .. ": " .. extra
  end

  return nil, msg, CLUSTERING_SYNC_STATUS.KONG_VERSION_INCOMPATIBLE
end


local function check_kong_version_compatibility(cp_version, dp_version)
  local cp = assert(parse_version(cp_version))
  local dp = assert(parse_version(dp_version))

  -- happy path
  if cp.major == dp.major and
     cp.minor == dp.minor
  then
    return true, nil, CLUSTERING_SYNC_STATUS.NORMAL

  -- if a minimum DP version is defined, use it
  elseif MIN_DP_VERSION and
         MIN_DP_VERSION > dp.number
  then
    return kong_incompatible(dp, cp)

  -- if no minimum DP version is defined, the DP major version must
  -- not be less than the CP major version
  elseif not MIN_DP_VERSION and
         dp.major < cp.major
  then
    return kong_incompatible(dp, cp)

  -- the DP should never be newer than the CP
  elseif (cp.major < dp.major) or
         (cp.major == dp.major and cp.minor < dp.minor)
  then
    return kong_incompatible(dp, cp, "data plane version cannot be higher than " ..
                                "the than control plane")

  -- not a happy path, but we can still try to sync
  elseif cp.minor ~= dp.minor then
    info("data plane minor version ", dp.minor,
         " != control plane minor version ", cp.minor)
  end

  return true, nil, CLUSTERING_SYNC_STATUS.NORMAL
end


_M.check_kong_version_compatibility = check_kong_version_compatibility


function _M.plugins_list_to_map(plugins_list)
  local versions = {}
  for _, plugin in ipairs(plugins_list) do
    local name = plugin.name
    local major, minor = extract_major_minor(plugin.version)

    if major and minor then
      versions[name] = {
        major   = major,
        minor   = minor,
        version = plugin.version,
      }

    else
      versions[name] = {}
    end
  end
  return versions
end


function _M.check_version_compatibility(cp, dp)
  return check_kong_version_compatibility(KONG_VERSION, dp.dp_version)
end


function _M.check_configuration_compatibility(cp, dp, conf)
  local ok, err, status = check_plugin_compatibility(cp, dp)
  if not ok then
    return nil, err, status
  end

  local dp_version = parse_version(dp.dp_version)

  if dp_version < parse_version("3.1") then
    for _, plugin in ipairs(conf.plugins or EMPTY) do
      local name = plugin.name
      local config = plugin.config or EMPTY

      -- check newly-added redis_ssl fields on acme and response-ratelimiting
      -- plugins
      do
        local redis_ssl
        if name == "acme" then
          redis_ssl = config.storage_config
                  and config.storage_config.redis
                  and config.storage_config.redis.ssl

        elseif name == "response-ratelimiting" then
          redis_ssl = config.redis_ssl
        end

        if redis_ssl == true then
          local msg = "redis ssl is not supported by data plane's version of the "
                      .. name ..  " plugin (" .. dp.dp_version .. ")"
          return nil, msg, CLUSTERING_SYNC_STATUS.PLUGIN_CONFIG_INCOMPATIBLE
        end
      end
    end
  end

  if dp_version < parse_version("3.0") then
    -- [[ XXX EE: Check for any WebSocket protocols
    --
    -- refactor me if/when adding new protocols becomes a regular thing
    local WS, WSS = "ws", "wss"

    local msg = "%s '%s' protocol '%s' is incompatible with data plane version %s"

    -- service.protocol
    local services = conf.services or EMPTY
    for _, service in ipairs(services) do
      local proto = service.protocol
      if proto == WS or proto == WSS then
        return nil,
               msg:format("service", service.id, proto, dp_version),
               CLUSTERING_SYNC_STATUS.SERVICE_PROTOCOL_INCOMPATIBLE
      end
    end

    -- route.protocols
    local routes = conf.routes or EMPTY
    for _, route in ipairs(routes) do
      if type(route.protocols) == "table" then
        for _, proto in ipairs(route.protocols) do
          if proto == WS or proto == WSS then
            return nil,
                   msg:format("route", route.id, proto, dp_version),
                   CLUSTERING_SYNC_STATUS.ROUTE_PROTOCOL_INCOMPATIBLE
          end
        end
      end
    end

    -- plugin.protocols
    local plugins = conf.plugins or EMPTY
    for _, plugin in ipairs(plugins) do
      if type(plugin.protocols) == "table" then
        local has_non_ws_proto = false
        local ws_proto = nil

        for _, proto in ipairs(plugin.protocols) do
          if proto == WS or proto == WSS then
            ws_proto = proto
          else
            has_non_ws_proto = true
          end
        end

        -- if the plugin has a mix of WS and non-WS protocols, we can handle
        -- that when updating the config payload later on
        --
        -- if the plugin _only_ has WS protocols, then it's probably not safe
        -- to continue
        if ws_proto and not has_non_ws_proto then
          return nil,
                 msg:format("plugin", plugin.id, ws_proto, dp_version),
                 CLUSTERING_SYNC_STATUS.PLUGIN_PROTOCOL_INCOMPATIBLE
        end
      end
    end
    -- XXX EE ]]

    local upstreams = conf.upstreams or EMPTY
    local msg = "upstream %s hash_on (%q) or hash_fallback (%q) is " ..
                "incompatible with data plane version %s"

    for _, upstream in ipairs(upstreams) do
      local hash_on = upstream.hash_on
      local hash_fallback = upstream.hash_fallback

      if hash_on == "path"
        or hash_on == "query_arg"
        or hash_on == "uri_capture"
        or hash_fallback == "path"
        or hash_fallback == "query_arg"
        or hash_fallback == "uri_capture"
      then
        return nil,
               msg:format(upstream.id, hash_on, hash_fallback, dp_version),
               CLUSTERING_SYNC_STATUS.UPSTREAM_HASH_INCOMPATIBLE
      end
    end
  end


  -- ]] EE

  return true, nil, CLUSTERING_SYNC_STATUS.NORMAL
end


local split_field_name
do
  local cache = {}

  --- a.b.c => { "a", "b", "c" }
  ---
  ---@param name string
  ---@return string[]
  function split_field_name(name)
    local fields = cache[name]
    if fields then
      return fields
    end

    fields = split(name, ".")

    for _, part in ipairs(fields) do
      assert(part ~= "", "empty segment in field name: " .. tostring(name))
    end

    cache[name] = fields
    return fields
  end
end


---@param t table
---@param key string
---@return boolean deleted
local function delete_at(t, key)
  local ref = t
  if type(ref) ~= "table" then
    return false
  end

  local addr = split_field_name(key)
  local len = #addr
  local last = addr[len]

  for i = 1, len - 1 do
    ref = ref[addr[i]]
    if type(ref) ~= "table" then
      return false
    end
  end

  if ref[last] ~= nil then
    ref[last] = nil
    return true
  end

  return false
end


local function invalidate_keys_from_config(config_plugins, keys, log_suffix)
  if not config_plugins then
    return false
  end

  local has_update

  for _, t in ipairs(config_plugins) do
    local config = t and t["config"]
    if config then
      local name = gsub(t["name"], "-", "_")

      if keys[name] ~= nil then
        for _, key in ipairs(keys[name]) do
          if delete_at(config, key) then
            warn(name, " plugin contains configuration '", key, "' which is ",
                 "incompatible with dataplane and will be ignored", log_suffix)
            has_update = true
          end
        end
      end
    end
  end

  return has_update
end


local get_removed_fields
do
  local cache = {}

  function get_removed_fields(dp_version)
    local dp = parse_version(dp_version)

    local plugin_fields = cache[dp.string]
    if plugin_fields ~= nil then
      return plugin_fields or nil
    end

    -- Merge dataplane unknown fields; if needed based on DP version
    for ver, plugins in pairs(REMOVED_FIELDS) do
      ver = parse_version(ver)
      if dp < ver then
        for plugin, items in pairs(plugins) do
          plugin_fields = plugin_fields or {}
          plugin_fields[plugin] = plugin_fields[plugin] or {}

          for _, name in ipairs(items) do
            table_insert(plugin_fields[plugin], name)
          end
        end
      end
    end

    if plugin_fields then
      -- sort for consistency
      for _, list in pairs(plugin_fields) do
        table_sort(list)
      end
      cache[dp.string] = plugin_fields
    else
      -- explicit negative cache
      cache[dp.string] = false
    end

    return plugin_fields
  end

  -- expose for unit tests
  _M._get_removed_fields = get_removed_fields
  _M._set_removed_fields = function(fields)
    local saved = REMOVED_FIELDS
    REMOVED_FIELDS = fields
    cache = {}
    return saved
  end
  _M._set_min_dp_version = function(s)
    local saved = MIN_DP_VERSION
    if type(s) == "string" then
      MIN_DP_VERSION = assert(parse_version(s))
    else
      MIN_DP_VERSION = s
    end
    return saved
  end
end

local function kafka_mechanism_compat(conf, plugin_name, dp_version, log_suffix)
  if plugin_name ~= "kafka-log" and plugin_name ~= "kafka-upstream" then
    return false
  end
  if conf["authentication"] then
    if conf["authentication"]["mechanism"] == "SCRAM-SHA-512" then
      warn("the kafka plugins for Kong Gateway v", KONG_VERSION, " contains ",
           "configuration mechanism='SCRAM-SHA-512', which is incompatible with ",
           "data plane version ", dp_version, " and will be set to ",
           "'SCRAM-SHA-256'.", log_suffix)
      conf["authentication"]["mechanism"] = "SCRAM-SHA-256"
      return true
    end
  end
  return false
end


-- returns has_update, modified_deflated_payload, err
function _M.update_compatible_payload(config, dp_version, log_suffix)
  local cp = KONG_VERSION
  local dp = parse_version(dp_version)

  -- if the CP and DP have the same version, avoid the payload
  -- copy and compatibility updates
  if dp.number == cp.number then
    return false
  end

  -- TODO: refactor refactor refactor:
  --
  -- 1. no excessive/unnecessary copying
  -- 2. avoid traversing the entire config multiple times
  config = deep_copy(config, false)
  local origin_config_table = utils.deep_copy(config, false)

  local has_update = false
  local fields = get_removed_fields(dp_version)

  if fields then
    has_update = invalidate_keys_from_config(config["plugins"], fields, log_suffix)
  end

  if dp < "3.1.0.0" then
    local config_upstream = config["upstreams"]
    if config_upstream then
      for _, t in ipairs(config_upstream) do
        if t["use_srv_name"] ~= nil then
          warn("Kong Gateway v", KONG_VERSION, " contains configuration ",
               "'upstream.use_srv_name' which is incompatible with dataplane ",
               " version ", dp_version, " and will be removed.", log_suffix)
          t["use_srv_name"] = nil
          has_update = true
        end
      end
    end
  end

  -- XXX EE: this should be moved in its own file (compat/config.lua). With a table
  -- similar to compat/remove_fields, each plugin could register a function to handle
  -- its compatibility issues.
  if dp < "3.0.0.0" then
    -- migrations for route path
    -- explicit regex indicator: https://github.com/Kong/kong/pull/9027
    -- no longer urldecode regex: https://github.com/Kong/kong/pull/9024
    if config._format_version == "3.0" then
      regex_router_migrate(config)
      config._format_version = "2.1"
      -- this is not evadiable cause we need to change the _format_version unconditionally
      has_update = true
    end

    if config["plugins"] then
      for i, t in ipairs(config["plugins"]) do
        local config = t and t["config"]
        if config then
          if t["name"] == "zipkin" then
            if config["header_type"] and config["header_type"] == "datadog" then
              warn("zipkin plugin for Kong Gateway v", KONG_VERSION,
                   " contains configuration 'header_type=datadog', which is incompatible with",
                   " dataplane version ", dp_version, " and will be replaced with 'preserve'.", log_suffix)
              config["header_type"] = "preserve"
              has_update = true
            end

            if config["default_header_type"] and config["default_header_type"] == "datadog" then
              warn("zipkin plugin for Kong Gateway v", KONG_VERSION,
                   " contains configuration 'default_header_type=datadog', which is incompatible with",
                   " dataplane version ", dp_version, " and will be replaced with 'b3'.", log_suffix)
              config["header_type"] = "b3"
              has_update = true
            end
          end

          if t["name"] == "statsd-advanced" then
            if config["metrics"] then
              local origin_config = origin_config_table["plugins"][i]["config"]
              for _, metric in ipairs(config["metrics"]) do
                warn("statsd-advanced plugin for Kong Gateway v", KONG_VERSION,
                     " supports null of consumer_identifier, service_identifier, and workspace_identifier,",
                     " which is incompatible with",
                     " dataplane version ", dp_version, " and will be replaced with the default value from",
                     " consumer_identifier_default, service_identifier_default, and workspace_identifier_default",
                     log_suffix)

                if not metric.consumer_identifier or metric.consumer_identifier == null then
                  metric.consumer_identifier = origin_config.consumer_identifier_default
                  has_update = true
                end

                if not metric.service_identifier or metric.service_identifier == null then
                  metric.service_identifier = origin_config.service_identifier_default
                  has_update = true
                end

                if not metric.workspace_identifier or metric.workspace_identifier == null then
                  metric.workspace_identifier = origin_config.workspace_identifier_default
                  has_update = true
                end
              end
            end
          end

          if t["name"] == "statsd" then
            local metrics = config["metrics"]
            if metrics then
              local origin_config = origin_config_table["plugins"][i]["config"]
              local removed_metrics = {
                status_count_per_workspace = true,
                status_count_per_user_per_route = true,
                shdict_usage = true,
                cache_datastore_hits_total = true,
                cache_datastore_misses_total = true,
              }

              for idx = #metrics, 1, -1 do
                local metric_name = metrics[idx]["name"]
                if removed_metrics[metric_name] then
                  warn("statsd plugin for Kong Gateway v", KONG_VERSION,
                       " supports metric '", metric_name, "' which is incompatible with",
                       " dataplane version ", dp_version, " and will be removed", log_suffix)
                  table.remove(metrics, idx)
                  has_update = true
                end
              end

              for _, metric in ipairs(metrics) do
                if not metric.consumer_identifier or metric.consumer_identifier == null then
                  metric.consumer_identifier = origin_config.consumer_identifier_default
                  has_update = true
                end

                if metric.service_identifier then
                  metric.service_identifier = nil
                  has_update = true
                end

                if metric.workspace_identifier then
                  metric.workspace_identifier = nil
                  has_update = true
                end
              end
            end
          end

          if t["name"] == "http-log" then
            if config["headers"] then
              -- no warning, because I only change the internal data type
              for header_name, header_value in pairs(config["headers"]) do
                local values = {}
                local parts = utils.split(header_value, ",")
                for _, v in ipairs(parts) do
                  table_insert(values, v)
                end
                config["headers"][header_name] = values
              end
              has_update = true
            end
          end
        end


        -- remove `ws` and `wss` from plugin.protocols if found
        local protocols = t and t["protocols"]
        if type(protocols) == "table" then
          local found_ws_proto

          for _, proto in ipairs(protocols) do
            if proto == "ws" or proto == "wss" then
              found_ws_proto = true
              break
            end
          end

          if found_ws_proto then
            warn(t["name"], " plugin for Kong Gateway",
                 " v", KONG_VERSION, " contains WebSocket protocols (ws/wss),",
                 " which are incompatible with dataplane version ", dp_version,
                 " and will be removed")

            has_update = true
            local new = {}
            for _, proto in ipairs(protocols) do
              if proto ~= "ws" and proto ~= "wss" then
                table_insert(new, proto)
              end
            end
            t["protocols"] = new
          end
        end
      end
    end

    if config["upstreams"] then
      -- handle new upstream `hash_on` and `hash_fallback` options
      -- https://github.com/Kong/kong/pull/8701

      local field_names = {
        "hash_on_query_arg",
        "hash_fallback_query_arg",
        "hash_on_uri_capture",
        "hash_fallback_uri_capture",
      }

      local removed = {}

      for _, t in ipairs(config["upstreams"]) do
        -- At this point the upstream's hash fields are have been validated as
        -- safe for pre-3.0 data planes, but it might still have incompatible
        -- fields present. Example scenario:
        --
        -- 1. POST /upstreams name=test
        --                    hash_on=query_arg
        --                    hash_on_query_arg=test
        --
        -- 2. PATCH /upstream/test hash_on=ip
        --
        -- The `hash_on` field is compatible with <3.0, but there is now a
        -- dangling `hash_on_query_arg` arg field that must be removed.

        local n = 0
        for _, field in ipairs(field_names) do
          if t[field] ~= nil then
            n = n + 1
            removed[n] = field
            t[field] = nil
          end
        end

        if n > 0 then
          local removed_field_names = table_concat(removed, ", ", 1, n)
          warn("Kong Gateway v", KONG_VERSION,
               " contains upstream configuration (", removed_field_names, ")",
               " which is incompatible with dataplane version ", dp_version,
               " and will be removed.", log_suffix)
          has_update = true
        end
      end
    end


  end

  if dp < "2.8.1.3" then
    -- [[ XXX EE: do not send vault entities to DP <2.8.1.3, there is a bug
    --            and no vault backends are enabled.
    if config.vaults then
      config.vaults = nil
      warn("vault backends of data plane version ",  dp_version,
           " are not compatible with version ", KONG_VERSION, "; it is strongly recommended to ",
           "upgrade your data plane to version ", KONG_VERSION)
      has_update = true
    end
    -- XXX EE ]]
  end


  if dp < "2.8.1.1" then
    if config["plugins"] then
      for _, t in ipairs(config["plugins"]) do
        local config = t and t["config"]
        if config then
          if kafka_mechanism_compat(config, t["name"], dp_version, log_suffix) then
            has_update = true
          end
        end
      end
    end
  end


  if has_update then
    return true, config
  end

  return false
end


return _M
