-- This software is copyright Kong Inc. and its licensors.
-- Use of the software is subject to the agreement between your organization
-- and Kong Inc. If there is no such agreement, use is governed by and
-- subject to the terms of the Kong Master Software License Agreement found
-- at https://konghq.com/enterprisesoftwarelicense/.
-- [ END OF LICENSE 0867164ffc95e54f04670b5169c09574bdbd9bba ]


local utils = require("kong.tools.utils")

local type = type
local tonumber = tonumber
local split = utils.split
local fmt = string.format
local getmetatable = getmetatable
local setmetatable = setmetatable

local MAJOR_MINOR_PATTERN = "^(%d+)%.(%d+)%.%d+"

local _M = {}


---@param  version     string
---@return integer|nil major
---@return integer|nil minor
function _M.extract_major_minor(version)
  if type(version) ~= "string" then
    return nil, nil
  end

  local major, minor = version:match(MAJOR_MINOR_PATTERN)
  if not major then
    return nil, nil
  end

  major = tonumber(major, 10)
  minor = tonumber(minor, 10)

  return major, minor
end


---@param s string
---@return integer
function _M.string_to_number(s)
  local base = 1000000000
  local num = 0
  for _, v in ipairs(split(s, ".", 4)) do
    v = v:match("^(%d+)")
    num = num + base * (tonumber(v, 10) or 0)
    base = base / 1000
  end

  return num
end

---@class kong.clustering.compat.version
---
---@field major      integer
---@field minor      integer
---@field patch      integer
---@field enterprise integer
---@field suffix     string|nil
---@field number     integer
---@field string     string

do
  local cache = setmetatable({}, { __mode = "v" })
  local _invalid = {}

  local function invalid(s)
    cache[s] = _invalid
    return nil, "invalid version string: " .. s
  end

  local mt

  local function is_version(t)
    return getmetatable(t) == mt
  end

  local from_string

  local function get_number(v)
    local t = type(v)

    if t == "number" then
      return v

    elseif t == "string" then
      local ver = from_string(v)
      if ver then
        return ver.number
      end

    elseif is_version(v) then
      return v.number
    end

    error("invalid version number: " .. tostring(v))
  end

  mt = {
    __tostring = function(self)
      return self.string
    end,

    __lt = function(self, other)
      return get_number(self) < get_number(other)
    end,

    __le = function(self, other)
      return get_number(self) <= get_number(other)
    end,

    __gt = function(self, other)
      return get_number(self) > get_number(other)
    end,

    __ge = function(self, other)
      return get_number(self) >= get_number(other)
    end,

    __eq = function(self, other)
      return get_number(self) == get_number(other)
    end,
  }


  ---@param s string
  ---@return kong.clustering.compat.version? version
  ---@return string? error
  function _M.from_string(s)
    local v = cache[s]

    if v == _invalid then
      return invalid(s)
    elseif v then
      return v
    end

    if is_version(s) then
      return s
    end

    local parts = split(s, ".", 4)
    local n = #parts

    -- for sanity's sake, we need to have at _least_ major+minor
    if n < 2 then
      cache[s] = invalid
      return invalid(s)
    end

    local major, minor, patch, ee, suffix

    major = tonumber(parts[1], 10)
    minor = tonumber(parts[2], 10)

    if n == 4 then
      patch = tonumber(parts[3], 10)

      ee, suffix = parts[4]:match("^([0-9]+)(.*)")
      ee = tonumber(ee, 10)

      suffix = suffix or ""

    elseif n == 3 then
      patch, suffix = parts[3]:match("^([0-9]+)(.*)")
      patch = tonumber(patch, 10)

      ee = 0

      suffix = suffix or ""

    elseif n == 2 then
      minor, suffix = parts[2]:match("^([0-9]+)(.*)")
      minor = tonumber(minor, 10)

      patch = 0
      ee = 0

      suffix = suffix or ""
    end

    if not major or not minor or not patch or not ee then
      return invalid(s)

    elseif major < 0 or major > 999
        or minor < 0 or minor > 999
        or patch < 0 or patch > 999
        or ee    < 0 or ee    > 999
    then
      return invalid(s)
    end

    local num = (major * 1000000000)
              + (minor * 1000000)
              + (patch * 1000)
              + (ee)

    local s_normalized = fmt("%d.%d.%d.%d%s",
                             major, minor,
                             patch, ee, suffix)


    local cached = cache[s_normalized]

    -- we found an equivalent version object in the cache,
    -- indexed under the normalized version string.
    --
    -- store the version object with the non-normalized string
    -- as well to memoize future lookups
    if cached then
      cache[s] = cached
      return cached
    end

    v = {
      major      = major,
      minor      = minor,
      patch      = patch,
      enterprise = ee,
      suffix     = suffix,
      number     = num,
      string     = s_normalized,
    }

    setmetatable(v, mt)

    cache[s_normalized] = v
    if s_normalized ~= s then
      cache[s] = v
    end

    return v
  end

  from_string = _M.from_string
end

return _M
